from pathlib import Path
from typing import Any
import json
from datetime import datetime, timedelta

from fastapi import APIRouter, Response
from fastapi.responses import StreamingResponse
from pydantic import BaseModel
import requests

from devcached.config import DevcachedConfig
from devcached.utils import download_binary_file, file_iterator


MIME_TYPE = "application/vnd.pub.v2+json"


class PubConfig(BaseModel):
    # Number of hours the cache is valid
    ttl: int = 7 * 24


def patch_pubdev_api_response(resp: dict[str, Any], package: str, base_url: str):
    # Patch the latest version
    if "latest" in resp:
        version = resp["latest"]["version"]
        resp["latest"]["archive_url"] = f"{base_url}/api/archives/{package}/{version}"

    # Patch all other versions
    new_versions = []
    for release in resp["versions"]:
        version = release["version"]
        release["archive_url"] = f"{base_url}/api/archives/{package}/{version}"
        new_versions.append(release)

    resp["versions"] = new_versions
    return resp


def create_module(config: DevcachedConfig) -> APIRouter:
    pub_config = PubConfig(**config.modules["devcached.modules.pub"])
    router = APIRouter(prefix="/pub/api")
    base_url = config.base_url + "/pub"
    data_dir = Path(config.data) / "pub"
    data_dir.mkdir(parents=True, exist_ok=True)

    @router.get("/packages/{package}")
    def get_package_info(package: str):
        package_dir = data_dir / package
        package_dir.mkdir(parents=True, exist_ok=True)
        package_info = package_dir / "package.json"
        ttl_file = package_dir / "ttl.txt"

        ttl: datetime | None = None
        if ttl_file.exists():
            with ttl_file.open("r", encoding="utf8") as f:
                ttl = datetime.fromisoformat(f.read())

        now = datetime.now()
        if not package_info.exists() or ttl is not None and now > ttl:
            resp = requests.get(
                f"https://pub.dev/api/packages/{package}",
                headers={
                    "Accept": MIME_TYPE,
                },
            )
            if resp.status_code == 200:
                payload = patch_pubdev_api_response(resp.json(), package, base_url)
                with package_info.open("w", encoding="utf8") as f:
                    json.dump(payload, f)
                with ttl_file.open("w", encoding="utf8") as f:
                    ttl = now + timedelta(hours=pub_config.ttl)
                    f.write(ttl.isoformat())
            else:
                payload = resp.json()
        else:
            with package_info.open("r", encoding="utf8") as f:
                payload = json.load(f)

        return Response(
            headers={
                "Content-Type": MIME_TYPE,
            },
            content=json.dumps(payload),
        )

    @router.get("/archives/{package}/{version}")
    def get_package_archive(package: str, version: str):
        package_dir = data_dir / package
        package_dir.mkdir(parents=True, exist_ok=True)
        version_file = package_dir / f"{version}.tar.gz"

        if not version_file.exists():
            download_binary_file(
                version_file,
                f"https://pub.dartlang.org/packages/{package}/versions/{version}.tar.gz",
            )

        return StreamingResponse(
            file_iterator(version_file),
            media_type="application/octet-stream",
        )

    return router
