import enum
from pathlib import Path
from datetime import datetime, timedelta
from typing import Any
import json
import logging
from typing import Annotated
from dataclasses import dataclass

from fastapi import APIRouter, Header, HTTPException, Response
from fastapi.responses import StreamingResponse
from pydantic import BaseModel
import requests

from devcached.config import DevcachedConfig
from devcached.utils import find, file_iterator, download_binary_file


class PypiConfig(BaseModel):
    # Number of hours the cache is valid.
    ttl: int = 7 * 24


class IndexRequestType(enum.Enum):
    JSON = enum.auto()


JSON_ACCEPT = "application/vnd.pypi.simple.v1+json"

logger = logging.getLogger("pypi")


def get_index_request_type(accept: str) -> IndexRequestType:
    if ";" in accept:
        accept = accept.split(";")[0]

    if "," in accept:
        parts = accept.split(",")
        accept = parts[0].strip()

    if accept == JSON_ACCEPT:
        return IndexRequestType.JSON
    else:
        raise Exception()


def _get_pypi_files(package: str) -> dict[str, Any]:
    logger.info("Requesting files for package %s", package)
    resp = requests.get(
        f"https://pypi.org/simple/{package}/",
        headers={
            "Accept": "application/vnd.pypi.simple.v1+json",
        },
    )
    return resp.json()


@dataclass
class PackageMetadata:
    ttl: datetime
    files: dict[str, Any]


def _download_files_json(
    config: PypiConfig,
    base_url: str,
    data_dir: Path,
    package: str,
) -> PackageMetadata:
    package_dir = data_dir / package
    package_dir.mkdir(parents=True, exist_ok=True)
    package_info = _get_pypi_files(package)

    # Patch the URLs so that we can proxy them
    package_files = [
        {
            **file,
            "original_url": file["url"],
            "url": f"{base_url}/pypi/proxy/{package}/{file['filename']}",
        }
        for file in package_info["files"]
    ]
    package_info["files"] = package_files

    with (package_dir / "files.json").open("w") as f:
        json.dump(package_info, f)

    # Write the TTL somewhere
    ttl = datetime.now() + timedelta(hours=config.ttl)
    with open(package_dir / "ttl.txt", "w") as f:
        f.write(ttl.isoformat())

    return PackageMetadata(
        ttl=ttl,
        files=package_info,
    )


def create_module(config: DevcachedConfig) -> APIRouter:
    pypi_config = PypiConfig(**config.modules["devcached.modules.pypi"])
    data_dir = Path(config.data) / "python"
    data_dir.mkdir(parents=True, exist_ok=True)
    router = APIRouter(prefix="/pypi")

    # TTL for packages
    package_ttl: dict[str, datetime] = {}

    @router.get("/simple/{package}/")
    def get_files(package: str, accept: Annotated[str, Header()]):
        logger.debug("Package index request using %s", accept)

        # Do we have the package?
        package_dir = data_dir / package
        now = datetime.now()

        # Read the TTL
        ttl_file = package_dir / "ttl.txt"
        if ttl_file.exists():
            with open(ttl_file, "r") as f:
                package_ttl[package] = datetime.fromisoformat(f.read().strip())

        if (
            not package_dir.exists()
            or package in package_ttl
            and now > package_ttl[package]
        ):
            metadata = _download_files_json(
                pypi_config,
                config.base_url,
                data_dir,
                package,
            )
            package_ttl[package] = metadata.ttl
            package_info = metadata.files
        else:
            logger.info("Answering files response from cache for %s", package)
            with (package_dir / "files.json").open("r") as f:
                package_info = json.load(f)

        index_type = get_index_request_type(accept)
        if index_type == IndexRequestType.JSON:
            logger.debug("Answering in JSON format")
            return Response(
                headers={
                    "Content-Type": JSON_ACCEPT,
                },
                content=json.dumps(package_info),
                status_code=200,
            )
        else:
            raise HTTPException(status_code=500)

    @router.get("/proxy/{package}/{filename}")
    def file_proxy(package: str, filename: str):
        package_dir = data_dir / package
        if not package_dir.exists():
            metadata = _download_files_json(
                pypi_config,
                config.base_url,
                data_dir,
                package,
            )
            package_ttl[package] = metadata.ttl

        file_path = package_dir / filename
        if not file_path.exists():
            logging.info("Downloading file for %s", package)
            with open(data_dir / package / "files.json", "r") as f:
                files = json.load(f)

            is_metadata_file = filename.endswith(".metadata")
            filename = filename.replace(".metadata", "")

            # Find original file's URL
            matching_file = find(
                files["files"],
                lambda file: file["filename"] == filename,
            )
            if matching_file is None:
                raise HTTPException(
                    detail="File not found",
                    status_code=404,
                )

            # Patch the URL in case the .metadata file got requested
            url = matching_file["original_url"]
            if is_metadata_file:
                url = url.split("#")[0] + ".metadata"
            download_binary_file(
                file_path,
                url,
            )

        return StreamingResponse(
            file_iterator(file_path),
            media_type="application/octet-stream",
        )

    return router
